package laboratoire4;

import java.util.Collection;

public class BoardEvaluator {
    public static int evaluate(MovingBoard game) {
        Collection<MovingPawn> maxPawns = game.getMaxPawns();
        Collection<MovingPawn> minPawns = game.getMinPawns();

        return evaluatePlayer(game, maxPawns, minPawns, true) - evaluatePlayer(game, minPawns, maxPawns, false);
    }

    //region Full Board
    private static int evaluatePlayer(MovingBoard game, Collection<MovingPawn> maxPawns, Collection<MovingPawn> minPawns, boolean max) {
        if (minPawns.stream().noneMatch(IPawn::isPusher)) {
            return 50;
        }

        int score = evaluateMaxPawnCountScore(maxPawns, minPawns);

        for (MovingPawn pawn : maxPawns) {
            score += evaluatePawn(game, pawn);

            if (max) {
                score += strategy(game, pawn);
            }
        }

        return score;
    }

    private static int evaluateMaxPawnCountScore(Collection<MovingPawn> maxPawns, Collection<MovingPawn> minPawns) {
        return evaluatePawnCountScore(minPawns) - evaluatePawnCountScore(maxPawns);
    }

    // Augmente beaucoup le score à partir de 10 pions capturés
    // f(x) = 0.001x^4, f(1) = 0, f(10) = 10, f(16) = 65.54
    private static int evaluatePawnCountScore(Collection<MovingPawn> pawns) {
        int capturedPusherCount = (int) (8 - pawns.stream().filter(IPawn::isPusher).count());
        return (int) (Math.pow(capturedPusherCount, 2));
    }
    //endregion

    //region Board
    private static int evaluatePawn(MovingBoard game, MovingPawn pawn) {
        if (pawn.getRow() == pawn.getPlayer().getGoal()) {
            return 50;
        }

        // Favorise les pièces qui peuvent gagner au prochain tour
        if (pawn.getRow() + pawn.getDirection() == pawn.getPlayer().getGoal()) {
            return 25;
        }

        int score = 0;

        score += evaluateHomePawnScore(pawn);
//        score += evaluatePushedScore(pawn);
        score += evaluatePusherScore(game, pawn);
        score += evaluateValidMovesScore(game, pawn);
        score += evaluateAttackScore(game, pawn);
        score += evaluateDefenseScore(game, pawn);

        if (Math.abs(pawn.getPlayer().getGoal() - pawn.getRow()) < 4) {
            score += hasEasyWinPath(game, pawn, pawn.getRow(), pawn.getCol(), 0);
        }

        return score;
    }

    private static int evaluateHomePawnScore(MovingPawn pawn) {
        return pawn.getRow() == pawn.getPlayer().getHome() ? 1 : 0;
    }

    private static int evaluatePushedScore(MovingPawn pawn) {
        return pawn.isPusher() ? 0 : 1;
    }

    private static int evaluatePusherScore(MovingBoard board, MovingPawn pawn) {
        if (!pawn.isPusher()) return 0;

        int score = 0;
        for (Pawn.PawnMovement movement : Pawn.PawnMovement.values()) {
            MovingPawn nearPawn = board.getNextPawn(pawn, movement);
            if (MovingPawn.areSamePlayers(pawn, nearPawn) && !nearPawn.isPusher()) {
                score++;
            }
        }
        return score;
    }

    private static int evaluateValidMovesScore(MovingBoard game, MovingPawn pawn) {
        int score = 0;
        for (Pawn.PawnMovement movement : Pawn.PawnMovement.values()) {
            if (pawn.isMoveValid(game.getBoard(), movement)) {
                score++;
            }
        }
        return score * pawn.getMoveCount();
    }

    private static int evaluateAttackScore(MovingBoard game, MovingPawn pawn) {
        int score = 0;

        Pawn.PawnMovement[] validAttackMoves = new Pawn.PawnMovement[]{Pawn.PawnMovement.LEFT_DIAGONAL, Pawn.PawnMovement.RIGHT_DIAGONAL};
        for (Pawn.PawnMovement movement : validAttackMoves) {
            MovingPawn nearPawn = game.getNextPawn(pawn, movement);
            if (nearPawn != null && pawn.isMoveValid(game.getBoard(), movement) && !MovingPawn.areSamePlayers(pawn, nearPawn)) {
                score++;
            }
        }

        // Favorise les attaques de notre côté
        if ((pawn.getPlayer() == Player.RED && pawn.getRow() < 4) ||
                (pawn.getPlayer() == Player.BLACK && pawn.getRow() >= 4)) {
            score++;
        }

        return score;
    }

    private static int evaluateDefenseScore(MovingBoard game, MovingPawn pawn) {
        int score = 0;

        Pawn.PawnMovement[] validAttackMoves = new Pawn.PawnMovement[]{Pawn.PawnMovement.LEFT_DIAGONAL, Pawn.PawnMovement.RIGHT_DIAGONAL};
        for (Pawn.PawnMovement movement : validAttackMoves) {
            MovingPawn nearPawn = game.getNextPawn(pawn, movement);
            Pawn.PawnMovement oppositeMovement = Pawn.PawnMovement.from(movement.getMove() * -1);

            if (nearPawn != null && !MovingPawn.areSamePlayers(pawn, nearPawn) && nearPawn.isMoveValid(game.getBoard(), oppositeMovement)) {
                score -= 50;
            }
        }

        return score;
    }

    private static int hasEasyWinPath(MovingBoard game, MovingPawn pawn, int row, int col, int depth) {
        if (row == pawn.getPlayer().getGoal()) {
            return 10;
        }

        int score = 0;
        int nextRow = row + pawn.getDirection();

        Pawn.PawnMovement[] preferredMoves = new Pawn.PawnMovement[]{Pawn.PawnMovement.LEFT_DIAGONAL, Pawn.PawnMovement.RIGHT_DIAGONAL};
        for (Pawn.PawnMovement movement : preferredMoves) {
            int nextCol = col + movement.getMove();
            if (nextCol < 0 || nextCol > 7) continue;

            MovingPawn nearPawn = game.getBoard()[nextRow][nextCol];

            if (nearPawn == null) {
                score += 2;
            } else if (nearPawn.getPlayer() != pawn.getPlayer()) {
                score++;
            } else {
                continue;
            }

            if (depth < 3) {
                score += hasEasyWinPath(game, pawn, nextRow, pawn.getCol() + movement.getMove(), depth + 1);
            }
        }

        return score;
    }
    //endregion

    //region Strategy
    private static int strategy(MovingBoard game, MovingPawn pawn) {
        int score = 0;

        score += defensiveStrategy(game, pawn);
        return score;
    }

    private static int defensiveStrategy(MovingBoard game, MovingPawn pawn) {
        int col = pawn.getCol();
        int row = pawn.getRow();

        if (col < 2 || col > 5) {
            return 0;
        }

        MovingPawn leftPawn = game.getBoard()[row][col - 2];
        MovingPawn rightPawn = game.getBoard()[row][col - 2];
        int score = 0;

        if (MovingPawn.areSamePlayers(leftPawn, pawn)) {
            score++;
        }
        if (MovingPawn.areSamePlayers(rightPawn, pawn)) {
            score++;
        }

        if (hasEnemyNear(game, pawn)) {
            score += 100;
        }

        return score;
    }

    private static boolean hasEnemyNear(MovingBoard game, MovingPawn pawn) {
        int row = pawn.getRow();
        int col = pawn.getCol();

        for (int i = 1; i <= 2; i++) {
            int nextRow = row + i * pawn.getDirection();
            if (nextRow < 0 || nextRow > 7) {
                break;
            }

            for (int j = -1; j <= 1; j++) {
                int nextCol = col + j;
                if (nextCol < 0 || nextCol > 7) {
                    continue;
                }

                MovingPawn nearPawn = game.getBoard()[nextRow][nextCol];
                if (nearPawn != null && !MovingPawn.areSamePlayers(pawn, nearPawn)) {
                    return true;
                }
            }
        }

        return false;
    }
    //endregion
}
